import concurrent.futures
import math
from typing import Dict, List, Protocol, runtime_checkable

import requests
from pydantic import StrictInt

from preemo.gen.endpoints.batch_create_artifact_part_pb2 import (
    BatchCreateArtifactPartRequest,
    CreateArtifactPartConfig,
    CreateArtifactPartConfigMetadata,
)
from preemo.gen.endpoints.batch_create_artifact_pb2 import (
    BatchCreateArtifactRequest,
    CreateArtifactConfig,
)
from preemo.gen.endpoints.batch_finalize_artifact_pb2 import (
    BatchFinalizeArtifactRequest,
    FinalizeArtifactConfig,
)
from preemo.gen.endpoints.batch_get_artifact_part_pb2 import (
    BatchGetArtifactPartRequest,
    GetArtifactPartConfig,
    GetArtifactPartConfigMetadata,
)
from preemo.gen.endpoints.batch_get_artifact_pb2 import (
    BatchGetArtifactRequest,
    GetArtifactConfig,
)
from preemo.worker._messaging_client import IMessagingClient
from preemo.worker._types import ImmutableModel, StringValue
from preemo.worker._validation import ensure_keys_match


class ArtifactId(StringValue):
    pass


class Artifact(ImmutableModel):
    id: ArtifactId
    part_size_threshold: StrictInt


@runtime_checkable
class IArtifactManager(Protocol):
    def create_artifact(self, content: bytes) -> ArtifactId:
        pass

    def create_artifacts(self, contents: List[bytes]) -> List[ArtifactId]:
        pass

    def get_artifact(self, artifact_id: ArtifactId) -> bytes:
        pass

    def get_artifacts(self, artifact_ids: List[ArtifactId]) -> List[bytes]:
        pass


class ArtifactManager:
    @staticmethod
    def _calculate_part_count(*, content_length: int, part_size_threshold: int) -> int:
        return max(1, math.ceil(content_length / part_size_threshold))

    @staticmethod
    def _upload_content(*, content: memoryview, url: str) -> None:
        response = requests.put(
            url=url, data=content, headers={"Content-Encoding": "gzip"}
        )

        # TODO(adrian@preemo.io, 04/15/2023): should retry if it fails
        if not response.ok:
            raise Exception(f"unexpected response while uploading: {response}")

    @staticmethod
    def _download_content(*, url: str) -> bytes:
        response = requests.get(url=url, headers={"Accept-Encoding": "gzip"})

        # TODO(adrian@preemo.io, 04/15/2023): should retry if it fails
        if not response.ok:
            raise Exception(f"unexpected response while downloading: {response}")

        return response.content

    def __init__(
        self,
        *,
        max_download_threads: int,
        max_upload_threads: int,
        messaging_client: IMessagingClient,
    ) -> None:
        if max_download_threads <= 0:
            raise ValueError("max_download_threads must be positive")

        if max_upload_threads <= 0:
            raise ValueError("max_upload_threads must be positive")

        self._max_download_threads = max_download_threads
        self._max_upload_threads = max_upload_threads
        self._messaging_client = messaging_client

    def _create_artifacts(self, *, count: int) -> List[Artifact]:
        configs_by_index = {i: CreateArtifactConfig() for i in range(count)}
        response = self._messaging_client.batch_create_artifact(
            BatchCreateArtifactRequest(configs_by_index=configs_by_index)
        )

        return [
            Artifact(
                id=ArtifactId(value=result.artifact_id),
                part_size_threshold=result.part_size_threshold,
            )
            for _, result in sorted(
                response.results_by_index.items(), key=lambda x: x[0]
            )
        ]

    def create_artifact(self, content: bytes) -> ArtifactId:
        artifact_ids = self.create_artifacts([content])
        if len(artifact_ids) != 1:
            raise Exception("expected exactly one artifact to be created")

        return artifact_ids[0]

    def create_artifacts(self, contents: List[bytes]) -> List[ArtifactId]:
        artifacts = self._create_artifacts(count=len(contents))
        if len(artifacts) != len(contents):
            raise Exception("expected artifacts and contents lengths to be equal")
        artifacts_and_contents = zip(artifacts, contents)

        configs_by_artifact_id_value: Dict[str, CreateArtifactPartConfig] = {}
        for artifact, content in artifacts_and_contents:
            part_count = ArtifactManager._calculate_part_count(
                content_length=len(content),
                part_size_threshold=artifact.part_size_threshold,
            )

            configs_by_artifact_id_value[artifact.id.value] = CreateArtifactPartConfig(
                metadatas_by_part_number={
                    part_number: CreateArtifactPartConfigMetadata()
                    for part_number in range(part_count)
                }
            )

        response = self._messaging_client.batch_create_artifact_part(
            BatchCreateArtifactPartRequest(
                configs_by_artifact_id=configs_by_artifact_id_value
            )
        )

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_upload_threads
        ) as executor:
            futures = []
            for artifact, content in artifacts_and_contents:
                content_view = memoryview(content)

                config = configs_by_artifact_id_value[artifact.id.value]
                result = response.results_by_artifact_id[artifact.id.value]
                ensure_keys_match(
                    expected=config.metadatas_by_part_number,
                    actual=result.metadatas_by_part_number,
                )

                for part_number, metadata in result.metadatas_by_part_number.items():
                    start_index = part_number * artifact.part_size_threshold
                    part_content = content_view[
                        start_index : start_index + artifact.part_size_threshold
                    ]

                    futures.append(
                        executor.submit(
                            ArtifactManager._upload_content,
                            content=part_content,
                            url=metadata.upload_signed_url,
                        )
                    )

            # TODO(adrian@preemo.io, 04/15/2023): add exception handling
            done, not_done = concurrent.futures.wait(
                futures, return_when=concurrent.futures.ALL_COMPLETED
            )

            if len(not_done) != 0:
                raise Exception("expected incomplete future set to be empty")

            if len(done) != len(futures):
                raise Exception("expected all futures to have completed")

        self._messaging_client.batch_finalize_artifact(
            BatchFinalizeArtifactRequest(
                configs_by_artifact_id={
                    artifact.id.value: FinalizeArtifactConfig(
                        total_size=len(content),
                        part_count=ArtifactManager._calculate_part_count(
                            content_length=len(content),
                            part_size_threshold=artifact.part_size_threshold,
                        ),
                    )
                    for artifact, content in artifacts_and_contents
                }
            )
        )

        return [artifact.id for artifact in artifacts]

    def get_artifact(self, artifact_id: ArtifactId) -> bytes:
        contents = self.get_artifacts([artifact_id])
        if len(contents) != 1:
            raise Exception("expected exactly one artifact to be retrieved")

        return contents[0]

    def get_artifacts(self, artifact_ids: List[ArtifactId]) -> List[bytes]:
        get_artifact_response = self._messaging_client.batch_get_artifact(
            BatchGetArtifactRequest(
                configs_by_artifact_id={
                    artifact_id.value: GetArtifactConfig()
                    for artifact_id in artifact_ids
                }
            )
        )

        configs_by_artifact_id_value = {
            artifact_id_value: GetArtifactPartConfig(
                metadatas_by_part_number={
                    part_number: GetArtifactPartConfigMetadata()
                    for part_number in range(result.part_count)
                }
            )
            for artifact_id_value, result in get_artifact_response.results_by_artifact_id.items()
        }
        get_artifact_part_response = self._messaging_client.batch_get_artifact_part(
            BatchGetArtifactPartRequest(
                configs_by_artifact_id=configs_by_artifact_id_value
            )
        )

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_download_threads
        ) as executor:
            futures_by_artifact_id_and_part_number: Dict[
                ArtifactId, Dict[int, concurrent.futures.Future]
            ] = {}
            for (
                artifact_id_value,
                artifact_part_result,
            ) in get_artifact_part_response.results_by_artifact_id.items():
                config = configs_by_artifact_id_value[artifact_id_value]
                ensure_keys_match(
                    expected=config.metadatas_by_part_number,
                    actual=artifact_part_result.metadatas_by_part_number,
                )

                futures_by_part_number: Dict[int, concurrent.futures.Future] = {}
                for (
                    part_number,
                    metadata,
                ) in artifact_part_result.metadatas_by_part_number.items():
                    futures_by_part_number[part_number] = executor.submit(
                        ArtifactManager._download_content,
                        url=metadata.download_signed_url,
                    )

                futures_by_artifact_id_and_part_number[
                    ArtifactId(value=artifact_id_value)
                ] = futures_by_part_number

            futures = [
                future
                for futures_by_part_number in futures_by_artifact_id_and_part_number.values()
                for future in futures_by_part_number.values()
            ]
            done, not_done = concurrent.futures.wait(
                futures,
                return_when=concurrent.futures.ALL_COMPLETED,
            )

            if len(not_done) != 0:
                raise Exception("expected incomplete future set to be empty")

            if len(done) != len(futures):
                raise Exception("expected all futures to have completed")

            results: List[bytes] = []
            for artifact_id in artifact_ids:
                futures_by_part_number = futures_by_artifact_id_and_part_number[
                    artifact_id
                ]
                get_artifact_result = get_artifact_response.results_by_artifact_id[
                    artifact_id.value
                ]

                # TODO(adrian@preemo.io, 04/20/2023): Consider other options for constructing the byte result,
                # such as a memory-mapped file or BytesIO
                result = bytearray()
                for part_number, future in sorted(
                    futures_by_part_number.items(), key=lambda x: x[0]
                ):
                    content = future.result()

                    if part_number < get_artifact_result.part_count - 1:
                        # not the final part
                        if len(content) != get_artifact_result.part_size_threshold:
                            raise Exception(
                                "expected content size to equal part_size_threshold"
                            )

                    result.extend(content)

                if len(result) != get_artifact_result.total_size:
                    raise Exception("expected result object size to equal total_size")

                # TODO(adrian@preemo.io, 04/20/2023): do i need to convert to bytes here? Is bytearray just masquerading as bytes?
                results.append(result)

        return results



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class GetArtifactConfig(google.protobuf.message.Message):
    """This config is included for potential future use."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetArtifactConfig = GetArtifactConfig

@typing_extensions.final
class BatchGetArtifactRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ConfigsByArtifactIdEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___GetArtifactConfig: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___GetArtifactConfig | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    CONFIGS_BY_ARTIFACT_ID_FIELD_NUMBER: builtins.int
    @property
    def configs_by_artifact_id(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___GetArtifactConfig]: ...
    def __init__(
        self,
        *,
        configs_by_artifact_id: collections.abc.Mapping[builtins.str, global___GetArtifactConfig] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["configs_by_artifact_id", b"configs_by_artifact_id"]) -> None: ...

global___BatchGetArtifactRequest = BatchGetArtifactRequest

@typing_extensions.final
class GetArtifactResult(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PART_COUNT_FIELD_NUMBER: builtins.int
    PART_SIZE_THRESHOLD_FIELD_NUMBER: builtins.int
    TOTAL_SIZE_FIELD_NUMBER: builtins.int
    part_count: builtins.int
    """Required field indicating the number of parts uploaded by the artifact."""
    part_size_threshold: builtins.int
    """Required field indicating the precise number of bytes each part contains,
    other than the final part, which can contain fewer bytes.
    """
    total_size: builtins.int
    """Required field indicating the size in bytes of the combined parts."""
    def __init__(
        self,
        *,
        part_count: builtins.int | None = ...,
        part_size_threshold: builtins.int | None = ...,
        total_size: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_part_count", b"_part_count", "_part_size_threshold", b"_part_size_threshold", "_total_size", b"_total_size", "part_count", b"part_count", "part_size_threshold", b"part_size_threshold", "total_size", b"total_size"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_part_count", b"_part_count", "_part_size_threshold", b"_part_size_threshold", "_total_size", b"_total_size", "part_count", b"part_count", "part_size_threshold", b"part_size_threshold", "total_size", b"total_size"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_part_count", b"_part_count"]) -> typing_extensions.Literal["part_count"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_part_size_threshold", b"_part_size_threshold"]) -> typing_extensions.Literal["part_size_threshold"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_total_size", b"_total_size"]) -> typing_extensions.Literal["total_size"] | None: ...

global___GetArtifactResult = GetArtifactResult

@typing_extensions.final
class BatchGetArtifactResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ResultsByArtifactIdEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___GetArtifactResult: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___GetArtifactResult | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    RESULTS_BY_ARTIFACT_ID_FIELD_NUMBER: builtins.int
    @property
    def results_by_artifact_id(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___GetArtifactResult]: ...
    def __init__(
        self,
        *,
        results_by_artifact_id: collections.abc.Mapping[builtins.str, global___GetArtifactResult] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["results_by_artifact_id", b"results_by_artifact_id"]) -> None: ...

global___BatchGetArtifactResponse = BatchGetArtifactResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import preemo.gen.endpoints.execute_function_pb2
import preemo.gen.endpoints.terminate_pb2
import grpc

class SDKServiceStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    ExecuteFunction: grpc.UnaryUnaryMultiCallable[
        preemo.gen.endpoints.execute_function_pb2.ExecuteFunctionRequest,
        preemo.gen.endpoints.execute_function_pb2.ExecuteFunctionResponse,
    ]
    Terminate: grpc.UnaryUnaryMultiCallable[
        preemo.gen.endpoints.terminate_pb2.TerminateRequest,
        preemo.gen.endpoints.terminate_pb2.TerminateResponse,
    ]

class SDKServiceServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def ExecuteFunction(
        self,
        request: preemo.gen.endpoints.execute_function_pb2.ExecuteFunctionRequest,
        context: grpc.ServicerContext,
    ) -> preemo.gen.endpoints.execute_function_pb2.ExecuteFunctionResponse: ...
    @abc.abstractmethod
    def Terminate(
        self,
        request: preemo.gen.endpoints.terminate_pb2.TerminateRequest,
        context: grpc.ServicerContext,
    ) -> preemo.gen.endpoints.terminate_pb2.TerminateResponse: ...

def add_SDKServiceServicer_to_server(servicer: SDKServiceServicer, server: grpc.Server) -> None: ...



import concurrent.futures
import random

import grpc

from preemo.gen.services.sdk_pb2_grpc import add_SDKServiceServicer_to_server
from preemo.worker._artifact_manager import IArtifactManager
from preemo.worker._function_registry import FunctionRegistry
from preemo.worker._sdk_service import SDKService


class SDKServer:
    @staticmethod
    def _generate_random_port() -> int:
        return random.randrange(60_000, 61_000)

    @staticmethod
    def _bind_server_to_random_port(*, server: grpc.Server, host: str) -> int:
        attempt_count = 0
        while True:
            port = SDKServer._generate_random_port()
            try:
                # TODO(adrian@preemo.io, 03/27/2023): investigate whether it makes sense to use add_secure_port instead
                server.add_insecure_port(f"{host}:{port}")
            except RuntimeError as e:
                if len(e.args) < 1:
                    raise e

                message = e.args[0]
                if "Failed to bind to address" not in message:
                    raise e

                print(f"failed to bind to port {port}, retrying with a different port")
            else:
                return port

            attempt_count += 1
            if attempt_count >= 20:
                raise Exception(f"failed to connect {attempt_count} times")

    def __init__(
        self,
        *,
        artifact_manager: IArtifactManager,
        function_registry: FunctionRegistry,
        sdk_server_host: str,
    ) -> None:
        server = grpc.server(
            concurrent.futures.ThreadPoolExecutor(max_workers=1),
            # This option prevents multiple servers from reusing the same port (see https://groups.google.com/g/grpc-io/c/RB69llv2tC4/m/7E__iL3LAwAJ)
            options=(("grpc.so_reuseport", 0),),
        )

        def close() -> None:
            server.stop(grace=10)  # seconds

        add_SDKServiceServicer_to_server(
            SDKService(
                artifact_manager=artifact_manager,
                function_registry=function_registry,
                terminate_server=close,
            ),
            server,
        )
        port = SDKServer._bind_server_to_random_port(
            server=server, host=sdk_server_host
        )

        server.start()
        print(f"sdk server has started on port {port}")

        self._server = server
        self._port = port

    def get_port(self) -> int:
        return self._port

    def wait_until_close(self) -> None:
        self._server.wait_for_termination()



import threading
from typing import Optional

from preemo.gen.endpoints.sdk_server_ready_pb2 import (
    SDKServerReadyRequest as _SDKServerReadyRequest,
)
from preemo.worker._artifact_manager import ArtifactManager as _ArtifactManager
from preemo.worker._artifact_manager import IArtifactManager as _IArtifactManager
from preemo.worker._env import get_optional_int_env as _get_optional_int_env
from preemo.worker._env import get_optional_string_env as _get_optional_string_env
from preemo.worker._function_registry import FunctionRegistry as _FunctionRegistry
from preemo.worker._messaging_client import IMessagingClient as _IMessagingClient
from preemo.worker._messaging_client import (
    LocalMessagingClient as _LocalMessagingClient,
)
from preemo.worker._messaging_client import MessagingClient as _MessagingClient
from preemo.worker._sdk_server import SDKServer as _SDKServer
from preemo.worker._worker_client import WorkerClient as _WorkerClient


def _construct_artifact_manager(
    messaging_client: _IMessagingClient,
) -> _IArtifactManager:
    max_download_threads = _get_optional_int_env("PREEMO_MAX_DOWNLOAD_THREADS")
    if max_download_threads is None:
        max_download_threads = 5

    max_upload_threads = _get_optional_int_env("PREEMO_MAX_UPLOAD_THREADS")
    if max_upload_threads is None:
        max_upload_threads = 5

    return _ArtifactManager(
        max_download_threads=max_download_threads,
        max_upload_threads=max_upload_threads,
        messaging_client=messaging_client,
    )


def _construct_messaging_client() -> _IMessagingClient:
    worker_server_url = _get_optional_string_env("PREEMO_WORKER_SERVER_URL")

    if worker_server_url is None:
        return _LocalMessagingClient()

    return _MessagingClient(worker_server_url=worker_server_url)


def _start_sdk_server(
    *, artifact_manager: _IArtifactManager, function_registry: _FunctionRegistry
) -> Optional[_SDKServer]:
    sdk_server_host = _get_optional_string_env("PREEMO_SDK_SERVER_HOST")
    if sdk_server_host is None:
        return None

    server = _SDKServer(
        artifact_manager=artifact_manager,
        function_registry=function_registry,
        sdk_server_host=sdk_server_host,
    )
    # This thread will keep the process running until the server is closed
    threading.Thread(target=lambda: server.wait_until_close()).start()

    return server


def _construct_worker_client() -> _WorkerClient:
    messaging_client = _construct_messaging_client()
    artifact_manager = _construct_artifact_manager(messaging_client=messaging_client)

    function_registry = _FunctionRegistry()
    sdk_server = _start_sdk_server(
        artifact_manager=artifact_manager, function_registry=function_registry
    )

    if sdk_server is not None:
        messaging_client.sdk_server_ready(
            _SDKServerReadyRequest(sdk_server_port=sdk_server.get_port())
        )

    return _WorkerClient(
        artifact_manager=artifact_manager,
        function_registry=function_registry,
        messaging_client=messaging_client,
    )


# provide shorthand for functions
__all__ = ["get_function", "parallel", "register"]

_worker_client = _construct_worker_client()

get_function = _worker_client.get_function
parallel = _worker_client.parallel
register = _worker_client.register



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: endpoints/batch_execute_function.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from models import registered_function_pb2 as models_dot_registered__function__pb2
from models import value_pb2 as models_dot_value__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&endpoints/batch_execute_function.proto\x12\tendpoints\x1a models/registered_function.proto\x1a\x12models/value.proto\"\x98\x02\n\x1b\x42\x61tchExecuteFunctionRequest\x12<\n\x13\x66unction_to_execute\x18\x01 \x01(\x0b\x32\x1a.models.RegisteredFunctionH\x00\x88\x01\x01\x12Z\n\x13parameters_by_index\x18\x02 \x03(\x0b\x32=.endpoints.BatchExecuteFunctionRequest.ParametersByIndexEntry\x1aG\n\x16ParametersByIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.models.Value:\x02\x38\x01\x42\x16\n\x14_function_to_execute\"\xbb\x01\n\x1c\x42\x61tchExecuteFunctionResponse\x12U\n\x10results_by_index\x18\x01 \x03(\x0b\x32;.endpoints.BatchExecuteFunctionResponse.ResultsByIndexEntry\x1a\x44\n\x13ResultsByIndexEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.models.Value:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'endpoints.batch_execute_function_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BATCHEXECUTEFUNCTIONREQUEST_PARAMETERSBYINDEXENTRY._options = None
  _BATCHEXECUTEFUNCTIONREQUEST_PARAMETERSBYINDEXENTRY._serialized_options = b'8\001'
  _BATCHEXECUTEFUNCTIONRESPONSE_RESULTSBYINDEXENTRY._options = None
  _BATCHEXECUTEFUNCTIONRESPONSE_RESULTSBYINDEXENTRY._serialized_options = b'8\001'
  _BATCHEXECUTEFUNCTIONREQUEST._serialized_start=108
  _BATCHEXECUTEFUNCTIONREQUEST._serialized_end=388
  _BATCHEXECUTEFUNCTIONREQUEST_PARAMETERSBYINDEXENTRY._serialized_start=293
  _BATCHEXECUTEFUNCTIONREQUEST_PARAMETERSBYINDEXENTRY._serialized_end=364
  _BATCHEXECUTEFUNCTIONRESPONSE._serialized_start=391
  _BATCHEXECUTEFUNCTIONRESPONSE._serialized_end=578
  _BATCHEXECUTEFUNCTIONRESPONSE_RESULTSBYINDEXENTRY._serialized_start=510
  _BATCHEXECUTEFUNCTIONRESPONSE_RESULTSBYINDEXENTRY._serialized_end=578
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: endpoints/sdk_server_ready.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n endpoints/sdk_server_ready.proto\x12\tendpoints\"I\n\x15SDKServerReadyRequest\x12\x1c\n\x0fsdk_server_port\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\x12\n\x10_sdk_server_port\"\x18\n\x16SDKServerReadyResponseb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'endpoints.sdk_server_ready_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SDKSERVERREADYREQUEST._serialized_start=47
  _SDKSERVERREADYREQUEST._serialized_end=120
  _SDKSERVERREADYRESPONSE._serialized_start=122
  _SDKSERVERREADYRESPONSE._serialized_end=146
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import preemo.gen.models.registered_function_pb2
import preemo.gen.models.value_pb2
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ExecuteFunctionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FUNCTION_TO_EXECUTE_FIELD_NUMBER: builtins.int
    PARAMETER_FIELD_NUMBER: builtins.int
    @property
    def function_to_execute(self) -> preemo.gen.models.registered_function_pb2.RegisteredFunction:
        """Required field specifying which function should be executed."""
    @property
    def parameter(self) -> preemo.gen.models.value_pb2.Value:
        """Required field representing the function parameter."""
    def __init__(
        self,
        *,
        function_to_execute: preemo.gen.models.registered_function_pb2.RegisteredFunction | None = ...,
        parameter: preemo.gen.models.value_pb2.Value | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_function_to_execute", b"_function_to_execute", "_parameter", b"_parameter", "function_to_execute", b"function_to_execute", "parameter", b"parameter"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_function_to_execute", b"_function_to_execute", "_parameter", b"_parameter", "function_to_execute", b"function_to_execute", "parameter", b"parameter"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_function_to_execute", b"_function_to_execute"]) -> typing_extensions.Literal["function_to_execute"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_parameter", b"_parameter"]) -> typing_extensions.Literal["parameter"] | None: ...

global___ExecuteFunctionRequest = ExecuteFunctionRequest

@typing_extensions.final
class ExecuteFunctionResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESULT_FIELD_NUMBER: builtins.int
    @property
    def result(self) -> preemo.gen.models.value_pb2.Value:
        """Required field representing the executed function result."""
    def __init__(
        self,
        *,
        result: preemo.gen.models.value_pb2.Value | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_result", b"_result", "result", b"result"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_result", b"_result", "result", b"result"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_result", b"_result"]) -> typing_extensions.Literal["result"] | None: ...

global___ExecuteFunctionResponse = ExecuteFunctionResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class SDKServerReadyRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SDK_SERVER_PORT_FIELD_NUMBER: builtins.int
    sdk_server_port: builtins.int
    """Required field for the port on which the SDK server is listening."""
    def __init__(
        self,
        *,
        sdk_server_port: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_sdk_server_port", b"_sdk_server_port", "sdk_server_port", b"sdk_server_port"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_sdk_server_port", b"_sdk_server_port", "sdk_server_port", b"sdk_server_port"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_sdk_server_port", b"_sdk_server_port"]) -> typing_extensions.Literal["sdk_server_port"] | None: ...

global___SDKServerReadyRequest = SDKServerReadyRequest

@typing_extensions.final
class SDKServerReadyResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___SDKServerReadyResponse = SDKServerReadyResponse



from typing import Callable, Optional

import grpc
from google.protobuf.struct_pb2 import NULL_VALUE

from preemo.gen.endpoints.execute_function_pb2 import (
    ExecuteFunctionRequest,
    ExecuteFunctionResponse,
)
from preemo.gen.endpoints.terminate_pb2 import TerminateRequest, TerminateResponse
from preemo.gen.models.value_pb2 import Value
from preemo.gen.services.sdk_pb2_grpc import SDKServiceServicer
from preemo.worker._artifact_manager import ArtifactId, IArtifactManager
from preemo.worker._function_registry import FunctionRegistry
from preemo.worker._types import assert_never


class SDKService(SDKServiceServicer):
    @staticmethod
    def _validate_execute_function_request(request: ExecuteFunctionRequest) -> None:
        if not request.HasField("function_to_execute"):
            raise Exception(
                "expected ExecuteFunctionRequest to have function_to_execute"
            )

        if not request.function_to_execute.HasField("name"):
            raise Exception("expected RegisteredFunction to have name")

        if not request.HasField("parameter"):
            raise Exception("expected ExecuteFunctionRequest to have parameter")

    def __init__(
        self,
        *,
        artifact_manager: IArtifactManager,
        function_registry: FunctionRegistry,
        terminate_server: Callable[[], None],
    ) -> None:
        self._artifact_manager = artifact_manager
        self._function_registry = function_registry
        self._terminate_server = terminate_server

    def _retrieve_value(self, value: Value) -> Optional[bytes]:
        kind = value.WhichOneof("kind")
        if kind is None:
            raise Exception("expected value to have kind")

        if kind == "null_value":
            return None

        if kind == "artifact_id":
            return self._artifact_manager.get_artifact(
                ArtifactId(value=value.artifact_id)
            )

        assert_never(kind)

    def ExecuteFunction(
        self, request: ExecuteFunctionRequest, context: grpc.ServicerContext
    ) -> ExecuteFunctionResponse:
        SDKService._validate_execute_function_request(request)

        if request.function_to_execute.HasField("namespace"):
            namespace = request.function_to_execute.namespace
        else:
            namespace = None

        func = self._function_registry.get_required_function(
            name=request.function_to_execute.name, namespace=namespace
        )

        parameter_value = self._retrieve_value(request.parameter)
        if parameter_value is None:
            result = func()
        else:
            result = func(parameter_value)

        if result is None:
            result_value = Value(null_value=NULL_VALUE)
        else:
            # TODO(adrian@preemo.io, 04/04/2023): validate that result is bytes?
            result_artifact_id = self._artifact_manager.create_artifact(result)
            result_value = Value(artifact_id=result_artifact_id.value)

        return ExecuteFunctionResponse(result=result_value)

    def Terminate(
        self, request: TerminateRequest, context: grpc.ServicerContext
    ) -> TerminateResponse:
        self._terminate_server()
        return TerminateResponse()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from endpoints import execute_function_pb2 as endpoints_dot_execute__function__pb2
from endpoints import terminate_pb2 as endpoints_dot_terminate__pb2


class SDKServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ExecuteFunction = channel.unary_unary(
                '/services.SDKService/ExecuteFunction',
                request_serializer=endpoints_dot_execute__function__pb2.ExecuteFunctionRequest.SerializeToString,
                response_deserializer=endpoints_dot_execute__function__pb2.ExecuteFunctionResponse.FromString,
                )
        self.Terminate = channel.unary_unary(
                '/services.SDKService/Terminate',
                request_serializer=endpoints_dot_terminate__pb2.TerminateRequest.SerializeToString,
                response_deserializer=endpoints_dot_terminate__pb2.TerminateResponse.FromString,
                )


class SDKServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ExecuteFunction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Terminate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SDKServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ExecuteFunction': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteFunction,
                    request_deserializer=endpoints_dot_execute__function__pb2.ExecuteFunctionRequest.FromString,
                    response_serializer=endpoints_dot_execute__function__pb2.ExecuteFunctionResponse.SerializeToString,
            ),
            'Terminate': grpc.unary_unary_rpc_method_handler(
                    servicer.Terminate,
                    request_deserializer=endpoints_dot_terminate__pb2.TerminateRequest.FromString,
                    response_serializer=endpoints_dot_terminate__pb2.TerminateResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'services.SDKService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SDKService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ExecuteFunction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/services.SDKService/ExecuteFunction',
            endpoints_dot_execute__function__pb2.ExecuteFunctionRequest.SerializeToString,
            endpoints_dot_execute__function__pb2.ExecuteFunctionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Terminate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/services.SDKService/Terminate',
            endpoints_dot_terminate__pb2.TerminateRequest.SerializeToString,
            endpoints_dot_terminate__pb2.TerminateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: endpoints/execute_function.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from models import registered_function_pb2 as models_dot_registered__function__pb2
from models import value_pb2 as models_dot_value__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n endpoints/execute_function.proto\x12\tendpoints\x1a models/registered_function.proto\x1a\x12models/value.proto\"\xa3\x01\n\x16\x45xecuteFunctionRequest\x12<\n\x13\x66unction_to_execute\x18\x01 \x01(\x0b\x32\x1a.models.RegisteredFunctionH\x00\x88\x01\x01\x12%\n\tparameter\x18\x02 \x01(\x0b\x32\r.models.ValueH\x01\x88\x01\x01\x42\x16\n\x14_function_to_executeB\x0c\n\n_parameter\"H\n\x17\x45xecuteFunctionResponse\x12\"\n\x06result\x18\x01 \x01(\x0b\x32\r.models.ValueH\x00\x88\x01\x01\x42\t\n\x07_resultb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'endpoints.execute_function_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXECUTEFUNCTIONREQUEST._serialized_start=102
  _EXECUTEFUNCTIONREQUEST._serialized_end=265
  _EXECUTEFUNCTIONRESPONSE._serialized_start=267
  _EXECUTEFUNCTIONRESPONSE._serialized_end=339
# @@protoc_insertion_point(module_scope)



from typing import Callable, List, Optional

class Result:
    def get(self) -> bytes: ...

class Function:
    def __call__(self, params: Optional[bytes] = None) -> Optional[Result]: ...

def get_function(name: str, *, namespace: Optional[str] = None) -> Function: ...
def parallel(
    function: Function,
    *,
    params: Optional[List[bytes]] = None,
    count: Optional[int] = None,
) -> List[Optional[Result]]: ...
def register(
    outer_function: Optional[Callable] = ...,
    *,
    name: Optional[str] = ...,
    namespace: Optional[str] = ...,
) -> Callable: ...
